﻿# CODE BLOCK BEGIN Global_Definitions 
from TSMaster import *
from Crypto.Hash import CMAC    # CMAC 계산에 필요하다.
from Crypto.Cipher import AES   # CMAC 계산의 알고리즘이다.
import random                   # 난수로 임의의 data를 만든다. 

# m_id
k_m_id_secoc_coord = 0x100
k_m_id_ecu_1 = 0x101

# CMAC을 사용한다.
# key는 128 비트, 16 바이트이다.
# 이 키는 비밀로 간직해야 한다.
# 실제로는 이렇게 코드에 노출하지 않는다.
k_key =  bytearray([0x00, 0x01, 0x02, 0x03, 0x04, 0x5, 0x06, 0x07, 0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f])

# 키 길이
k_len_key = len(k_key)  # Length of key

# CMAC 길이. 128 비트, 16 "바이트"다.
k_len_cmac = 16 # Length of CMAC

# CMAC_trunc의 길이
# CMAC 16 바이트 중에서 앞 2 바이트를 사용한다.
k_len_cmac_trunc = 2 # Length of CMAC truncated

# Freshness Value의 길이
# fv를 6 바이트로 한다.  
k_len_fv = 6 # Length of fv

# fv_trunc의 길이
# fv 6 바이트 중에 1 바이트를 사용한다.
k_len_fv_trunc = 1 # Length of fv truncated

# MAC(Message Authentication Code)의 길이
# fv_trunc 1 바이트와 cmac_trunc 2 바이트 해서 3 바이트이다.
k_len_mac = k_len_fv_trunc + k_len_cmac_trunc # Length of MAC

# CAN/-FD의 최대 메시지 길이
# python은 메지지 길이에 맞게 동적으로 저장 공간을 확보할 수 있지만
# TSMaster를 보니, 실제 메시지 길이와 무관하게 
# 모든 메시지에 64 바이트 저장 공간을 확보한다. 
# 이렇게 하는 것이 어떤 이점이 있는지 알아보려고 따라해본다. 
k_len_data_max = 64 # Maximum data length

# fv의 최대값
# fv가 6 바이트이면, fv의 최대값은 0xFF가 6개 이다.
k_fv_max = int(bytes([0xFF] * k_len_fv).hex(), 16)   
k_fv_init = 0x010203040506  # 6 바이트 숫자이다. 
k_fv_init_bytearray = k_fv_init.to_bytes(k_len_fv, byteorder='big')

k_m_dlc_ecu_1 = 0x0a
k_m_len_ecu_1 = 16

tx_msgs = {}
tx_msgs[k_m_id_ecu_1] = {
    'fv': 0,
}
tx_msgs[k_m_id_secoc_coord] = {
    'fv': 0,
}


# ecu_1 메시지를 정의한다.
msg_ecu_1 = RawCAN(
    k_m_id_ecu_1,  # id 
    k_m_dlc_ecu_1, # dlc 
    CH1,           # channel
    0,             # 모른다. 
    [i for i in range(k_m_len_ecu_1)], # 데이터. 리스트이다.  
    0x306,         # 설정. CAN-FD 메시지의 경우
)

k_m_dlc_secoc_coord = 0x0a

# secoc_coord 메시지를 정의한다.
# secoc_coord 메시지는 SecOC 관련 통신 관리에 사용된다.
# 예) 추적 중인 fv값을 fv_init으로 재설정한다.
# request, m_id_target, fv_init
msg_secoc_coord = RawCAN(
    k_m_id_secoc_coord, 
    k_m_dlc_secoc_coord, 
    CH1, 
    0, 
    [0x01, 0x01, 0x01, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06], 
    0x306
)  

k_pos_fv_trunc_from_eod = 3   # fv는 data의 끝에서 3번째에 있다.
k_pos_cmac_trunc_from_eod = 2 # cmac은 data의 끝에서 2번째부터 끝까지이다.


# function
def dlc_to_length(dlc, is_can_fd):
    '''
    아래 표에 따라 dlc(data length code) 값을 길이로 변환한다.
    CAN의 경우 dlc는 길이이다.
    CAN-FD의 경우, dlc는 길이가 아니다.
    
    dlc can can-fd
    0	0	0
    1	1	1
    2	2	2
    3	3	3
    4	4	4
    5	5	5
    6	6	6
    7	7	7
    8	8	8
    9	8	12
    10	8	16
    11	8	20
    12	8	24
    13	8	32
    14	8	48
    15	8	64
    '''
    length = [0, 1, 2, 3, 4, 5, 6, 7, 8, 12, 16, 20, 24, 32, 48, 64]

    if is_can_fd:
        return length[dlc]
    else:
        return min(length[dlc], 8)

# CODE BLOCK END Global_Definitions 
# CODE BLOCK BEGIN Instance 
Instance = MpInstance('SecOC_master')
# CODE BLOCK END Instance 
# CODE BLOCK BEGIN Timer timer_1 MTAwMA__
timer_1 = MpTimer("SecOC_master", "timer_1", 1000) # Auto generated code, for modification please goto python code editor
# CODE BLOCK END Timer timer_1
# CODE BLOCK BEGIN On_Shortcut ctrl_1
def on_shortcut_ctrl_1() -> None:
    app.log_text('timer_1 started', lvlOK)
    timer_1.Start()

# CODE BLOCK END On_Shortcut ctrl_1
# CODE BLOCK BEGIN On_Shortcut ctrl_0
def on_shortcut_ctrl_0() -> None:
    app.log_text('timer_1 stopped', lvlOK)
    timer_1.Stop()

# CODE BLOCK END On_Shortcut ctrl_0
# CODE BLOCK BEGIN On_Shortcut ctrl_5
def on_shortcut_ctrl_5() -> None:
    # request를 fv 재설정을 위한 0x01로 설정한다.
    msg_secoc_coord.set_data(0, 0x01)


    # m_id_target을 설정한다.
    # fv를 변경할 메시지의 아이디이다.
    m_id_target = k_m_id_ecu_1
    m_id_target_bytearray = m_id_target.to_bytes(2, byteorder='big')
    msg_secoc_coord.set_data(1, m_id_target_bytearray[0])
    msg_secoc_coord.set_data(2, m_id_target_bytearray[1])
    

    # fv_init을 설정한다.
    # 예제에서는 fv_init 값을 고정값으로 한다.
    
    
    # 전체 메시지의 fv를 변경할 때
    # m_id_target == 0xAAAA 
    
    if m_id_target == 0xAAAA:
        for tx_msg in tx_msgs:
            tx_msg['fv'] = k_fv_init 
    else:
        tx_msgs[m_id_target]['fv'] = k_fv_init


    # msg를 전송한다.
    if com.transmit_can_async(msg_secoc_coord):
        app.log_text('transmit failed', lvlInfo)
    else:
        fv = bytearray(msg_secoc_coord.data[3:9])
        app.log_text(f'reset {fv=}', lvlOK)

# CODE BLOCK END On_Shortcut ctrl_5
# CODE BLOCK BEGIN On_Timer timer_1 dGltZXJfMQ__
def on_timer_timer_1() -> None:
    global msg_ecu_1
    '''
    전송할 메시지 ecu_1을 준비한다.
        난수로 데이터를 채운다.
        fv_trunc를 데이터에 추가한다.
        cmac_trunc를 데이터에 추가한다. 
    메시지를전송한다. 
    '''
        
    global tx_msgs


    # msg에 msg_ecu_1을 할당한다.
    # 메새지가 여러 개 있을 때를 대비해서
    msg = msg_ecu_1    

    m_id = msg.id

    # 메시지 길이
    m_len = dlc_to_length(msg.dlc, msg.is_edl)

    # fv_trunc, cmac_trunc를 포함한 메시지 데이터를 난수로 채운다.
    msg.data = bytearray(random.randbytes(m_len))

    # 데이터에서 fv_trunc와 cmac_trunc를 제외한 
    # cmac 계산에 사용할 부분만 추출한다. 
    data = bytearray(msg.data[:m_len - k_len_mac]) 
    
    # fv를 증가한다.
    fv = tx_msgs[m_id]['fv']
    fv += 1
    # if fv_int > 0xFF FF FF FF FF FF: fv를 6 바이트로 했다. 
    if fv > k_fv_max:
        fv = 0

    # fv_trunc를 추출한다.
    fv_trunc_calc = bytearray.fromhex(f'{fv:0{k_len_fv * 2}x}')[-1]

    # msg.data에 fv_trunc를넣는다.
    pos_fv_trunc = m_len - k_pos_fv_trunc_from_eod
    msg.set_data(pos_fv_trunc, fv_trunc_calc)

    # cmac을 구한다.
    cmac_obj = CMAC.new(k_key, ciphermod=AES)
    cmac_obj.update(data)

    # fv는 int형이다. fv를 항상 일정한 길이로 하여 계산에 사용하기 위해 bytearray로 형변환한다.
    fv_bytearray = bytearray.fromhex(f'{fv:0{k_len_fv * 2}x}')    
    cmac_obj.update(fv_bytearray)

    cmac = cmac_obj.digest()

    # cmac_trunc를 구한다.    
    cmac_trunc = cmac[:k_len_cmac_trunc]

    # msg.data에 cmac_trunc를 넣는다.
    pos_cmac_trunc = m_len - k_pos_cmac_trunc_from_eod
    msg.set_data(pos_cmac_trunc, cmac_trunc[0])
    msg.set_data(pos_cmac_trunc + 1, cmac_trunc[1])

    # msg를 전송한다.
    if com.transmit_can_async(msg):
        app.log_text('transmit failed', lvlInfo)

    # 다음 메시지 전송 때 사용하기 위해 fv를 저장한다.
    tx_msgs[m_id]['fv'] = fv

# CODE BLOCK END On_Timer timer_1
# CODE BLOCK BEGIN Step_Function  MA__
def step() -> None:
    pass

# CODE BLOCK END Step_Function 
